const std = @import("std");
const sdl = @cImport({
    @cInclude("SDL3/SDL.h");
});
const Action = @import("../actions.zig").Action;
const IOInterface = @import("io_interface.zig").IOInterface;

const TEXT_SPRITE_INDEX = 0;
const TEXT_WIDTH = 10;
const TEXT_HEIGHT = 16;
const TEXT_SCALE = 2;

const SCREEN_WIDTH = 640;
const SCREEN_HEIGHT = 480;

pub const IO = struct {
    allocator: std.mem.Allocator,
    screen: ?*sdl.SDL_Window,
    screen_surface: ?*sdl.SDL_Surface,
    sprites: std.ArrayList(?*sdl.SDL_Surface),

    x_pos: c_int,
    y_pos: c_int,

    fn loadSprite(self: *IO) !void {
        const sprite = sdl.SDL_LoadBMP("assets/ascii.bmp");
        if (sprite == null) {
            std.log.err("{s}", .{sdl.SDL_GetError()});
            return error.FailedToLoadSprite;
        }

        const optimized_sprite = sdl.SDL_ConvertSurface(sprite, self.screen_surface.?.format);
        if (optimized_sprite == null) {
            std.log.err("{s}", .{sdl.SDL_GetError()});
            return error.FailedToOptimizeSprite;
        }

        _ = sdl.SDL_DestroySurface(sprite);
        try self.sprites.append(optimized_sprite);
    }

    pub fn displayMessage(self: *IO, str: []const u8) anyerror!void {
        // TODO: remove this
        _ = sdl.SDL_ClearSurface(self.screen_surface, 1, 1, 1, 1);

        for (str) |c| {
            if (c < ' ') continue;

            if (self.x_pos + TEXT_SCALE * TEXT_WIDTH >= SCREEN_WIDTH) {
                self.x_pos = 0;
                self.y_pos += TEXT_SCALE * TEXT_HEIGHT;
            } else {
                self.x_pos += TEXT_SCALE * TEXT_WIDTH;
            }

            const index: c_int = c - ' ';

            std.log.debug("index: 0x{x}", .{index});

            const src_rect = sdl.SDL_Rect{
                .x = index * TEXT_WIDTH,
                .y = 0,
                .w = TEXT_WIDTH,
                .h = TEXT_HEIGHT,
            };

            const dst_rect = sdl.SDL_Rect{
                .x = self.x_pos,
                .y = self.y_pos,
                .w = TEXT_SCALE * TEXT_WIDTH,
                .h = TEXT_SCALE * TEXT_HEIGHT,
            };

            if (!sdl.SDL_BlitSurfaceScaled(self.sprites.items[TEXT_SPRITE_INDEX], &src_rect, self.screen_surface, &dst_rect, sdl.SDL_SCALEMODE_NEAREST)) {
                return error.SDL_BlitSurfaceFailed;
            }
        }

        if (!sdl.SDL_UpdateWindowSurface(self.screen)) {
            return error.SDL_UpdateSurfaceFailed;
        }
    }

    pub fn waitForTick(self: *IO) anyerror!Action {
        _ = self;

        var quit = false;
        var event: sdl.SDL_Event = undefined;
        while (!quit) {
            _ = sdl.SDL_WaitEvent(&event);
            switch (event.type) {
                sdl.SDL_EVENT_QUIT => quit = true,
                else => {},
            }
        }

        return Action.exit;
    }

    pub fn init(allocator: std.mem.Allocator) !IOInterface {
        var io = try allocator.create(IO);
        io.allocator = allocator;
        io.sprites = std.ArrayList(?*sdl.SDL_Surface).init(allocator);

        io.x_pos = 0;
        io.y_pos = 0;

        if (!sdl.SDL_Init(sdl.SDL_INIT_VIDEO)) {
            std.log.err("{s}", .{sdl.SDL_GetError()});
            return error.SDL_Failed;
        }

        io.screen = sdl.SDL_CreateWindow(
            "urlg",
            SCREEN_WIDTH,
            SCREEN_HEIGHT,
            0,
        );

        if (io.screen == null) {
            std.log.err("{s}", .{sdl.SDL_GetError()});
            return error.SDL_Failed;
        }

        io.screen_surface = sdl.SDL_GetWindowSurface(io.screen);
        io.loadSprite() catch |err| {
            io.deinit();
            return err;
        };

        try io.displayMessage(" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghjklmnopqrstuvwxyz{|}~");
        //try io.displayMessage("THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG.");

        //_ = sdl.SDL_BlitSurfaceScaled(io.sprites.items[0], null, io.screen_surface, &rect, sdl.SDL_SCALEMODE_NEAREST);
        //rect.x += HEX_SCALE + 2 * HEX_ANGLE_DIST;
        //_ = sdl.SDL_BlitSurfaceScaled(io.sprites.items[0], null, io.screen_surface, &rect, sdl.SDL_SCALEMODE_NEAREST);

        if (!sdl.SDL_UpdateWindowSurface(io.screen)) {
            io.deinit();
            return error.SDL_UpdateSurfaceFailed;
        }

        return IOInterface.init(io);
    }

    pub fn deinit(self: *IO) void {
        for (self.sprites.items) |sprite| {
            _ = sdl.SDL_DestroySurface(sprite);
        }
        self.sprites.deinit();

        _ = sdl.SDL_DestroyWindow(self.screen);

        _ = sdl.SDL_Quit();
        self.allocator.destroy(self);
    }
};
